﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace zcompiler {
    class Scanner {

        //Token information
        private readonly IList<object> result;
        public IList<object> Tokens {
            get { return this.result; }
        }

        /// <summary>
        /// Default constructor
        /// </summary>
        public Scanner(TextReader input) {

            this.result = new List<object>();
            this.Scan(input);

        }

        /// <summary>
        /// Constants for arithmetic tokens
        /// </summary>
        public enum Ops {

            Add = 0,
            Sub = 1,
            Mult = 2,
            Div = 3,
            Power = 4,
            Mod = 5,
            Equal = 6,
            Smcln = 7,

        }

        private char[] escapeItems = new char[] { 'a', 'b', 'f', 'n', 'r', 't', 'v', '"', '\\', '0' };

        /// <summary>
        /// Breaks ASCII code down into tokens
        /// </summary>
        private void Scan(TextReader input) {

            //While we are within the input source
            while (input.Peek() != -1) { 
            
                //Get the next available character from the input
                char chr = (char)input.Peek();

                //Scan for every possible token
                if (char.IsWhiteSpace(chr)) {

                    //First, if the character is essentially nothing,
                    //then we're going to want to skip it entirely.
                    //Includes: \0, '   ' (aka \t)
                    input.Read();

                }
                else if (char.IsLetter(chr) || chr.Equals('_')) {
                    #region Char or letter check
                    //If the character is a letter, it could be
                    //a token or a variable, so we need to
                    //get its complete data
                    StringBuilder data = new StringBuilder();

                    //While reading the data (can only be a letter
                    //or an underscore, '_')...
                    while (char.IsLetter(chr) || chr.Equals('_')) {

                        //Add the character to the data
                        data.Append(chr);
                        //'Consume' that character
                        input.Read();

                        //If we can't read anymore, stop accumulating data
                        if (input.Peek() == -1)
                            break;
                        //Otherwise, read and accumulate data
                        else
                            chr = (char)input.Peek();

                    }

                    //Add the data to the token result
                    this.result.Add(data.ToString());
                    #endregion
                }
                else if (chr.Equals('"')) {
                    #region String check
                    //If we encounter a quote, then we've reached a string
                    //and we need to accumulate its data like a token or var
                    StringBuilder data = new StringBuilder();

                    //Skip the beginning "
                    input.Read();

                    //If we've reached the end of the source
                    if (input.Peek() == -1)
                        //We can't finish the string
                        throw new Exception("Unterminated string detected.");

                    //While we're in the string
                    while ((chr = (char)input.Peek()) != '"') {

                        //If the character is the escape char, aka \
                        if (chr.Equals('\\')) {

                            //Get the next character
                            char next = (char)input.Peek();
                            //Setup a flag to check if the next char is an escape char
                            bool isEsc = false;

                            //Go through each escape character
                            foreach (char esc in escapeItems)
                                //If the next char is an escape char
                                if (next == esc)
                                    //Set the flag to true
                                    isEsc = true;

                            //If the next character is not an escape char,
                            if (!isEsc)
                                //Throw an error
                                throw new Exception("Char '" + next + "' is not an escape char.");

                            //If the next character is an escape char,
                            //then set the parent char to the corresponding escape char
                            switch (next) {

                                case 'a':
                                    chr = '\a';
                                    break;
                                case 'b':
                                    chr = '\b';
                                    break;
                                case 'f':
                                    chr = '\f';
                                    break;
                                case 'n':
                                    chr = '\n';
                                    break;
                                case 'r':
                                    chr = '\r';
                                    break;
                                case 't':
                                    chr = '\t';
                                    break;
                                case 'v':
                                    chr = '\v';
                                    break;
                                case '"':
                                    chr = '"';
                                    break;
                                case '\\':
                                    chr = '\\';
                                    break;
                                case '0':
                                    chr = '\0';
                                    break;
                            
                            }

                        }

                        //Add the string data
                        data.Append(chr);
                        //'Consume' the current character
                        input.Read();

                        //Again, if we've reached the end of the source
                        if (input.Peek() == -1)
                            //We can't finish the string
                            throw new Exception("Unterminated string detected.");

                        //Skip the terminating quote
                        input.Read();
                        //Store the string data
                        this.result.Add(data);
                    }
                    #endregion
                }
                else if (char.IsDigit(chr) || chr.Equals('$')) {
                    //We can check for the $ because that will only appear
                    //for bytes and in strings, which we already check for

                    #region Number and byte check
                    //If the character is a number or byte, we need
                    //to get the entire number
                    StringBuilder data = new StringBuilder();

                    //While we're reading a number or a byte
                    while (char.IsDigit(chr) ||
                        chr.Equals('x') || chr.Equals('h')) {

                        //Add the char to data
                        data.Append(chr);
                        //'Consume' character and read on
                        input.Read();

                        //If we can't read anymore,
                        if (input.Peek() == -1)
                            //Exit
                            break;
                        //Otherwise
                        else
                            //Read next char
                            chr = (char)input.Peek();

                    }

                    //Now we just need to check if the data is a number or byte
                    bool isByte = false;
                    foreach (char c in data.ToString())
                        if ("$xh".Contains(c.ToString()))
                            isByte = true;

                    //If it's a byte, add a byte
                    if (isByte)
                        //I realize this will keep going if someone just does 0x0x0x0x0x0x0x0x...
                        //However, with this, we'll let VS catch the exception for us ;)
                        this.result.Add(byte.Parse(data.ToString()));
                    //If it's a number, add a number
                    else
                        this.result.Add(int.Parse(data.ToString()));

                    #endregion
                }
                else switch (chr) {
                    #region Other characters
                        //Pretty self explanatory IMO
                    case '+':
                        input.Read();
                        this.result.Add(Ops.Add);
                        break;

                    case '-':
                        input.Read();
                        this.result.Add(Ops.Sub);
                        break;

                    case '*':
                        input.Read();
                        this.result.Add(Ops.Mult);
                        break;

                    case '/':
                        input.Read();
                        this.result.Add(Ops.Div);
                        break;

                    case '=':
                        input.Read();
                        this.result.Add(Ops.Equal);
                        break;
                    
                    case ';':
                        input.Read();
                        this.result.Add(Ops.Smcln);
                        break;
                   
                    default:
                        throw new System.Exception(
                            String.Format("'{0}' is an unrecognized character.", chr));
                    #endregion
                }
            }
        }
    }
}
